import threading
import time
import can

_TX_MODE = 0
_TX_MODE__STOP = 0  # global, local
_TX_MODE__TX_CONT = 1  # global, local
_TX_MODE__PER_PGN = 2  # global only
_TX_MODE__TX_ONCE = 3  # local only
_TX_RATE_SEC = 1
_DATA = 2


class Transmitter:
    def __init__(self, J1939: dict, bus: can.BusABC):
        self._J1939 = J1939
        self._bus = bus
        self.PGNs_pending_tx = {}

        self.global_tx_mode = _TX_MODE__STOP  # _TX_MODE__STOP, _TX_MODE__TX_CONT, _TX_MODE__PER_PGN
        self.tx_CAN_IDs = {}  # {(CAN_ID: int, is_extended: bool): {
                              #     _TX_MODE: _TX_MODE__STOP, _TX_MODE__TX_CONT, _TX_MODE__TX_ONCE
                              #     _TX_RATE_SEC: #,
                              #     _DATA: list
                              # }

        self.J1939_CAN_IDs = {}  # {PGN: CAN_ID}

    def register_tx_PGN(self, *, pgn, priority, source_address, tx_rate_ms):
        can_id = priority << 26 | pgn << 8 | source_address
        self.J1939_CAN_IDs[pgn] = (can_id, True)
        tx_rate_sec = tx_rate_ms / 1000
        self.tx_CAN_IDs[(can_id, True)] = {
            _TX_MODE: _TX_MODE__STOP,
            _TX_RATE_SEC: tx_rate_sec,
            _DATA: [0 for _ in range(self._J1939[pgn]['PGN Data Length'])]
        }
        threading.Thread(target=self.send_periodic, args=(can_id, True), daemon=True).start()

    def send_periodic(self, can_id, is_extended):
        signal_spec = self.tx_CAN_IDs[(can_id, is_extended)]

        while True:
            time.sleep(signal_spec[_TX_RATE_SEC])
            if self.global_tx_mode == _TX_MODE__TX_CONT or signal_spec[_TX_MODE] == _TX_MODE__TX_CONT:
                pass
            elif self.global_tx_mode in [_TX_MODE__STOP, _TX_MODE__PER_PGN] \
                    and signal_spec[_TX_MODE] == _TX_MODE__TX_ONCE:
                signal_spec[_TX_MODE] = _TX_MODE__STOP
            else:
                continue

            self._bus.send(can.Message(arbitration_id=can_id, is_extended_id=is_extended, data=signal_spec[_DATA]))

    def set_tx_mode_stop(self, pgn=None):
        if pgn is None:
            self.global_tx_mode = _TX_MODE__STOP
        else:
            self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_TX_MODE] = _TX_MODE__STOP

    def set_tx_mode_continuous(self, pgn=None):
        if pgn is None:
            self.global_tx_mode = _TX_MODE__TX_CONT
        else:
            self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_TX_MODE] = _TX_MODE__TX_CONT

    def set_tx_mode_per_PGN(self):
        self.global_tx_mode = _TX_MODE__PER_PGN

    def set_tx_once(self, pgn=None):
        if pgn is None:
            for signal_spec in self.tx_CAN_IDs.values():
                signal_spec[_TX_MODE] = _TX_MODE__TX_ONCE
        else:
            self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_TX_MODE] = _TX_MODE__TX_ONCE

    def modify_pgn_tx_rate(self, pgn, tx_rate_ms):
        tx_rate_sec = tx_rate_ms / 1000
        self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_TX_RATE_SEC] = tx_rate_sec

    def modify_pgn_data(self, pgn: int, spn: int, raw_value: int):
        data = self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_DATA]
        spn_spec = self._J1939[pgn]['SPNs'][spn]
        start_byte = spn_spec['start_byte']

        match spn_spec['length_bits']:
            case 1 | 2 | 3 | 4 | 5 | 6 | 7:
                start_bit = spn_spec['start_bit']
                length_bits = spn_spec['length_bits']
                mask = 0x00
                for i in range(length_bits):
                    mask = mask | 1 << start_bit + i
                data[start_byte] = data[start_byte] & ~mask | raw_value << start_bit & mask
            case 8:
                data[start_byte] = raw_value
            case 16:
                data[start_byte] = raw_value & 0x00FF
                data[start_byte + 1] = raw_value >> 8
            case 32:
                data[start_byte] = raw_value & 0x0000_00FF
                data[start_byte + 1] = raw_value >> 8 & 0x00_00FF
                data[start_byte + 2] = raw_value >> 16 & 0x00FF
                data[start_byte + 3] = raw_value >> 24

        self.tx_CAN_IDs[self.J1939_CAN_IDs[pgn]][_DATA] = data



import platform
import os
import subprocess
from pathlib import Path
import tomllib
import pickle

import can

from .import J1939

default_config_toml = '''[CAN_INTERFACE]
# akrocansim uses the python-can library for utilising CAN interfaces.
# Any configuration parameters defined by python-can, can be listed here
# in the same format at they would appear in can.interface.Bus().
# See https://python-can.readthedocs.io/en/v4.3.1/interfaces.html

# See example key/value pairs below for PCAN-USB.

interface='pcan'
channel='PCAN_USBBUS1'
bitrate=250000


[J1939DA]
# Obtain a copy of the J1939 Digital Annex from SAE International at: https://www.sae.org/standards/?search=j1939DA
# The J1939 Digital Annex is a Microsoft Excel file.
# If your copy is in .xls format, you need to convert it to .xlsx using Microsoft Excel.
# Save the 'J1939DA_??????.xlsx' file in the 'J1939' folder, located in the same folder with this configuration.
# Edit the values of all keys in the [J1939DA] table as required and parse the J1939 Digital Annex.

filename = 'J1939DA_??????.xlsx'
SPNs_and_PGNs_sheet = 'SPNs & PGNs'


[J1939DA.SPNs_and_PGNs_sheet_columns]
# Adjust as required

'PGN' = 'E'
'Parameter Group Label' = 'F'
'Acronym' = 'G'
'PGN Description' = 'H'
'PGN Data Length' = 'O'
'Default Priority' = 'P'
'Transmission Rate' = 'N'
'SPN' = 'S'
'SPN Name' = 'T'
'SPN Description' = 'U'
'SPN Position in PGN' = 'R'
'SPN Length' = 'V'
'Resolution' = 'W'
'Offset' = 'X'
'Data Range' = 'Y'
'Operational Range' = 'Z'
'Units' = 'AA'


[J1939DA.SPNs_to_parse]
# Specify the range of rows in the J1939 Digital Annex SPNs_and_PGNs_sheet that will be scanned for SPN definitions.

first_row = 5
last_row = 5000


[Tx_PGNs_SPNs]
# List the PGNs and SPNs to be loaded by akrocansim using the following format:
# PGN = [SPN#1, SPN#2, ..., SPN#N], e.g. 61444 = [513, 190]

#61444 = [513, 190]
'''


class Config:
    def __init__(self, config_dir: Path = None):
        if config_dir is None:
            self.config_dir = Path.home()/'akrocansim'
        else:
            self.config_dir = config_dir
        self.config_toml = self.config_dir/'config.toml'
        self.J1939_dir = self.config_dir/'J1939'
        self.J1939DA_xlsx = None
        self.J1939_pickle = self.J1939_dir/'J1939.pkl'

        self._config = {}

        self.J1939_spec = None
        self.bus = None
        self.tx_PGNs_SPNs = {}

    def load(self) -> (str, list[str]):
        status_message = ''
        error_messages = []

        state = 'INIT'
        while True:
            match state:
                case 'INIT':
                    if self.config_toml.exists():
                        state = 'CONFIG_PRESENT'
                    else:
                        state = 'CONFIG_NOT_PRESENT'

                case 'CONFIG_NOT_PRESENT':
                    self.config_dir.mkdir(exist_ok=True)
                    self.J1939_dir.mkdir(exist_ok=True)
                    with self.config_toml.open(mode='w', encoding='utf-8') as f:
                        f.write(default_config_toml)
                    status_message = f'new configuration file created at {self.config_toml}'
                    state = 'RETURN'

                case 'CONFIG_PRESENT':
                    try:
                        with self.config_toml.open('rb') as f:
                            self._config = tomllib.load(f)
                        state = 'CHECK_CAN_CONFIG'
                    except tomllib.TOMLDecodeError as e:
                        status_message = 'configuration file cannot be loaded'
                        error_messages.append(f'FORMAT ERROR: {e}')
                        state = 'RETURN'

                case 'CHECK_CAN_CONFIG':
                    if self.bus is not None:
                        self.bus.shutdown()
                    try:
                        if not self._config['CAN_INTERFACE']:
                            status_message = 'incorrect or incomplete configuration'
                            error_messages.append('MISSING ELEMENT ERROR: CAN interface configuration parameters'
                                                  ' not found in [CAN_INTERFACE] section')
                        else:
                            try:
                                self.bus = can.Bus(**self._config['CAN_INTERFACE'])
                            except can.exceptions.CanError as e:
                                status_message = 'incorrect or incomplete configuration'
                                error_messages.append(f'CAN INTERFACE ERROR: {e}')
                    except KeyError:
                        status_message = 'incorrect or incomplete configuration'
                        error_messages.append(f'MISSING ELEMENT ERROR: [CAN_INTERFACE] section not found')
                    state = 'CHECK_J1939DA'

                case 'CHECK_J1939DA':
                    if '?' in self._config['J1939DA']['filename']:
                        status_message = 'incorrect or incomplete configuration'
                        error_messages.append(f'INCORRECT ELEMENT ERROR: [J1939DA] filename has not been set')
                        state = 'RETURN'
                    elif Path(self._config['J1939DA']['filename']).suffix != '.xlsx':
                        status_message = 'incorrect or incomplete configuration'
                        error_messages.append('INCORRECT ELEMENT ERROR: J1939DA filename must be in .xlsx format')
                        state = 'RETURN'
                    else:
                        self.J1939DA_xlsx = self.J1939_dir/self._config['J1939DA']['filename']
                        if not self.J1939DA_xlsx.exists():
                            status_message = 'incorrect or incomplete configuration'
                            error_messages.append(f'INCORRECT ELEMENT ERROR: {self.J1939DA_xlsx} file not found')
                            state = 'RETURN'
                        else:
                            state = 'CHECK_J1939_PKL'

                    # I think it's all good until here.

                case 'CHECK_J1939_PKL':
                    if not self.J1939_pickle.exists():
                        status_message = 'J1939DA has not been parsed'
                        state = 'RETURN'
                    else:
                        with self.J1939_pickle.open('rb') as f:
                            self.J1939_spec = pickle.load(f)
                            state = 'CHECK_TX_SPNS'

                case 'CHECK_TX_SPNS':
                    if not self._config['Tx_PGNs_SPNs']:
                        status_message = 'incorrect or incomplete configuration'
                        error_messages.append('INCORRECT ELEMENT ERROR: PGNs not found in [Tx_PGNs_SPNs] section')
                    else:
                        pgn_spn_errors_found = False
                        for pgn, spn_list in self._config['Tx_PGNs_SPNs'].items():
                            try:
                                _ = self.J1939_spec[int(pgn)]
                                for spn in spn_list:
                                    try:
                                        _ = self.J1939_spec[int(pgn)]['SPNs'][spn]
                                    except KeyError:
                                        pgn_spn_errors_found = True
                                        status_message = 'incorrect or incomplete configuration'
                                        error_messages.append(
                                            f'MISSING SPN ERROR: SPN {spn} not found in parsed elements of J1939DA')
                            except KeyError:
                                pgn_spn_errors_found = True
                                status_message = 'incorrect or incomplete configuration'
                                error_messages.append(
                                    f'MISSING PGN ERROR: PGN {pgn} not found in parsed elements of J1939DA')
                            if not pgn_spn_errors_found:
                                self.tx_PGNs_SPNs[int(pgn)] = spn_list
                            pgn_spn_errors_found = False

                    state = 'RETURN'

                case 'RETURN':
                    return status_message, error_messages

                case _:
                    raise Exception()

    def parse_J1939DA(self) -> str:
        parsing_result = J1939.parse_J1939DA(J1939_dir=self.J1939_dir, J1939DA_config=self._config['J1939DA'])
        return parsing_result

    def ext_browse(self) -> None:
        path = str(self.config_dir)
        if platform.system() == 'Windows':
            os.startfile(path)
        else:
            subprocess.Popen(["xdg-open", path])

    def ext_edit(self) -> None:
        filename = str(self.config_toml)
        if platform.system() == 'Windows':
            os.system(filename)
        else:
            os.system('%s %s' % (os.getenv('EDITOR'), filename))



import webbrowser
import os

import dearpygui.dearpygui as dpg

from .__init__ import __version__, __app_name__
from .transmitter import Transmitter
from .config import Config
from . import J1939

VIEWPORT_WIDTH = 1500
VIEWPORT_HEIGHT = 650
WINDOW_WIDTH = VIEWPORT_WIDTH - 16


def _hyperlink(text, address):
    b = dpg.add_button(label=text, callback=lambda: webbrowser.open(address))
    dpg.bind_item_theme(b, "__demo_hyperlinkTheme")

class AkrocansimGui:
    def __init__(self):
        self.config = Config()
        self.transmitter = None
        self.J1939: dict = None

        self.gui_main()

    def gui_main(self):
        dpg.create_context()
        icon = os.path.join(os.path.dirname(__file__), 'resources', 'akrocansim.ico')
        dpg.create_viewport(title=__app_name__, width=VIEWPORT_WIDTH, height=VIEWPORT_HEIGHT, vsync=True,
                            small_icon=icon, large_icon=icon)

        with dpg.theme(tag="__demo_hyperlinkTheme"):
            with dpg.theme_component(dpg.mvButton):
                dpg.add_theme_color(dpg.mvThemeCol_Button, [0, 0, 0, 0])
                dpg.add_theme_color(dpg.mvThemeCol_ButtonActive, [0, 0, 0, 0])
                dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, [29, 151, 236, 25])
                dpg.add_theme_color(dpg.mvThemeCol_Text, [29, 151, 236])

        logo = os.path.join(os.path.dirname(__file__), 'resources', 'akrocansim_logo_dark.png')
        logo_width, logo_height, logo_channels, logo_data = dpg.load_image(logo)
        with dpg.texture_registry():
            dpg.add_static_texture(logo_width, logo_height, logo_data, tag='logo')

        self.make_menu_bar()
        self.load_config()

        dpg.setup_dearpygui()
        dpg.show_viewport()
        dpg.start_dearpygui()

        dpg.destroy_context()
        if self.config.bus is not None:
            self.config.bus.shutdown()

    def make_menu_bar(self):
        with dpg.viewport_menu_bar():
            with dpg.menu(label="Settings"):
                dpg.add_menu_item(tag='folder_explore', label='Explore configuration folder', callback=self.config.ext_browse)
                dpg.add_menu_item(tag='config_edit', label='Edit configuration', callback=self.config.ext_edit)
                dpg.add_menu_item(tag='J1939_parse', label='Parse J1939 Digital Annex', callback=self.parse_J1939DA)
                dpg.add_menu_item(tag='config_load', label='Load configuration', callback=self.load_config)
            with dpg.menu(label='Help'):
                dpg.add_menu_item(label='Talk to us',
                                  callback=lambda: webbrowser.open('https://github.com/cfsok/akrocansim/discussions'))
                dpg.add_menu_item(label='Report a problem',
                                  callback=lambda: webbrowser.open('https://github.com/cfsok/akrocansim/issues'))
                dpg.add_menu_item(label='About', callback=self.make_about_window)
            dpg.add_text('  |  ')
            dpg.add_text(tag='status_bar')

    def make_about_window(self):
        with dpg.window(label=f'About {__app_name__}', pos=(200, 100), modal=True, no_resize=True, no_move=True, autosize=True):
            with dpg.group(horizontal=True):
                with dpg.group():
                    dpg.add_text('\n\n\n\n\nakrocansim\n'
                                 f'Version {__version__}\n'
                                 'Copyright 2023 Socrates Vlassis\n\n')
                    _hyperlink('Homepage (PyPI)', 'https://pypi.org/project/akrocansim/')
                dpg.add_spacer(width=20)
                dpg.add_image('logo')
            dpg.add_separator()
            dpg.add_spacer(height=20)
            dpg.add_text('akrocansim is make possible by the following open source projects:\n')
            _hyperlink('python-can', 'https://github.com/hardbyte/python-can')
            _hyperlink('DearPyGui', 'https://github.com/hoffstadt/DearPyGui')
            _hyperlink('openpyxl', 'https://openpyxl.readthedocs.io/')

    def load_config(self):
        status_message, error_messages = self.config.load()

        dpg.set_value(item='status_bar', value=status_message)
        if 'new configuration file created' in status_message:
            instructions = 'Use the Settings menu to edit the configuration file.'
            self.show_messages([instructions])
        elif 'J1939DA has not been parsed' in status_message:
            instructions = 'Use the Settings menu to parse J1939DA and reload configuration.'
            self.show_messages([instructions])
        elif error_messages:
            if dpg.does_item_exist('J1939_window'):
                dpg.delete_item('J1939_window')
            self.show_messages(error_messages)
        else:
            self.J1939 = self.config.J1939_spec
            self.transmitter = Transmitter(self.J1939, self.config.bus)
            self.make_PGN_global_tx_management_window()
            self.make_J1939_window()

    def parse_J1939DA(self):
        parsing_result = self.config.parse_J1939DA()
        dpg.set_value(item='status_bar', value=parsing_result)

    def show_messages(self, messages: list[str]):
        if not dpg.does_item_exist('config_messages'):
            with dpg.window(pos=(0, 19), width=WINDOW_WIDTH, height=VIEWPORT_HEIGHT - 58,
                            no_move=True, no_resize=True, no_title_bar=True, no_close=True):
                dpg.add_text(tag='config_messages')
        error_text = ''
        for message in messages:
            error_text += message + '\n'*2
        dpg.set_value(item='config_messages', value=error_text)

    def make_PGN_global_tx_management_window(self):  # put user_data='all PGNs' to tag
        with dpg.window(label=self.config.bus.channel_info, width=570, no_close=True):
            with dpg.group(horizontal=True):
                dpg.add_text('Continuous J1939 PGN transmission:')
                dpg.add_radio_button(tag='global_tx_mode', items=('Stop All', 'Tx All', 'Use PGN Settings'),
                                     default_value='Stop All', horizontal=True,
                                     callback=self.global_tx_mode_changed)
            dpg.add_button(tag='global_tx_once', label='Tx All PGNs Once', callback=self.global_tx_once_invoked)

    def global_tx_mode_changed(self, sender, tx_mode, user_data):
        match tx_mode:
            case 'Stop All':
                dpg.show_item('global_tx_once')
                self.transmitter.set_tx_mode_stop()
            case 'Tx All':
                dpg.hide_item('global_tx_once')
                self.transmitter.set_tx_mode_continuous()
            case 'Use PGN Settings':
                dpg.show_item('global_tx_once')
                self.transmitter.set_tx_mode_per_PGN()

    def global_tx_once_invoked(self, sender, app_data, user_data):
        self.transmitter.set_tx_once()

    def make_J1939_window(self):
        with dpg.window(tag='J1939_window', label='J1939', width=WINDOW_WIDTH, pos=(0, 119), height=500, no_close=True):
            for pgn, spns in self.config.tx_PGNs_SPNs.items():
                self.add_pgn(pgn, spns)

    def add_pgn(self, pgn, spns: list):
        priority = self.J1939[pgn]['Default Priority']
        source_address = 0
        self.transmitter.register_tx_PGN(pgn=pgn, priority=priority, source_address=source_address,
                                         tx_rate_ms=self.J1939[pgn]['transmission_rate_ms'])

        pgn_label = (f"PGN {pgn} - CAN ID: {priority << 26 | pgn << 8 | source_address:08X} "
                     f"- {self.J1939[pgn]['Acronym']} - {self.J1939[pgn]['Parameter Group Label']}")
        with (dpg.collapsing_header(label=pgn_label, default_open=True)):

            continuous_signals_present = False
            enum_signals_present = False

            for spn in spns:
                spn_spec = self.J1939[pgn]['SPNs'][spn]
                if spn_spec['scale'] not in ['ENUM']:
                    continuous_signals_present = True
                if spn_spec['scale'] == 'ENUM':
                    enum_signals_present = True

            if continuous_signals_present:
                with dpg.table(header_row=True, resizable=True,
                               borders_outerV=True, borders_outerH=True, borders_innerH=True, borders_innerV=True):
                    dpg.add_table_column(label='REAL VALUE', width_fixed=True)
                    dpg.add_table_column(label='UNIT', width_fixed=True)
                    dpg.add_table_column(label='RAW HEX', width_fixed=True)
                    dpg.add_table_column(label='RAW DECIMAL')

                    for spn in spns:
                        spn_spec = self.J1939[pgn]['SPNs'][spn]
                        if spn_spec['scale'] not in ['ENUM']:
                            with dpg.table_row():
                                # SIGNAL - decoded value, direct entry, decrement, decrement
                                max_value = (2 ** spn_spec['length_bits'] - 1) * spn_spec['scale'] + spn_spec['offset']
                                if type(spn_spec['scale']) is int:
                                    dpg.add_input_int(tag=f'{pgn}_{spn}_input', width=140,
                                                      user_data=(pgn, spn, spn_spec),
                                                      default_value=spn_spec['min_value'],
                                                      min_value=spn_spec['offset'],
                                                      max_value=max_value,
                                                      min_clamped=True, max_clamped=True,
                                                      step=spn_spec['scale'],
                                                      callback=self.continuous_spn_input_int_changed,
                                                      on_enter=True)
                                elif type(spn_spec['scale']) is float:
                                    dpg.add_input_float(tag=f'{pgn}_{spn}_input', width=140,
                                                        user_data=(pgn, spn, spn_spec),
                                                        format=f"%.{spn_spec['n_decimals']}f",
                                                        default_value=spn_spec['min_value'],
                                                        min_value=spn_spec['offset'],
                                                        max_value=max_value,
                                                        min_clamped=True, max_clamped=True,
                                                        step=spn_spec['scale'],
                                                        callback=self.continuous_spn_input_int_changed,
                                                        on_enter=True)
                                else:
                                    raise TypeError

                                # UNIT
                                unit = spn_spec['unit']
                                dpg.add_text(f"{unit if unit is not None else ''}".ljust(10))

                                # RAW HEX
                                dpg.add_text(tag=f'{spn}_hex')

                                # RAW DECIMAL
                                min_value = J1939.raw_min_value(signal_spec=spn_spec)
                                dpg.add_slider_int(tag=str(spn), user_data=(pgn, spn, spn_spec),
                                                   label=f"SPN {spn}: {spn_spec['SPN Name']}",
                                                   min_value=min_value,
                                                   max_value=J1939.raw_max_value(signal_spec=spn_spec),
                                                   default_value=J1939.raw_min_value(signal_spec=spn_spec),
                                                   callback=self.continuous_spn_slider_changed)
                            self.continuous_spn_slider_changed(str(spn), min_value, (pgn, spn, spn_spec))

            if enum_signals_present:
                with dpg.table(header_row=True, resizable=True,
                               borders_outerV=True, borders_outerH=True, borders_innerH=True, borders_innerV=True):
                    dpg.add_table_column(label='DECIMAL', width_fixed=True)
                    dpg.add_table_column(label='BINARY', width_fixed=True)
                    dpg.add_table_column(label='HEX', width_fixed=True)
                    dpg.add_table_column(label='OPTIONS')

                    for spn in spns:
                        spn_spec = self.J1939[pgn]['SPNs'][spn]
                        if spn_spec['scale'] == 'ENUM':
                            with dpg.table_row():
                                # DECIMAL
                                dpg.add_input_int(tag=str(spn), width=140, user_data=(pgn, spn, spn_spec),
                                                  min_value=0, max_value=2 ** spn_spec['length_bits'] - 1,
                                                  min_clamped=True, max_clamped=True,
                                                  callback=self.discrete_spn_input_int_changed,
                                                  on_enter=True)

                                # BINARY
                                dpg.add_text(tag=f'{spn}_bin')

                                # HEX
                                dpg.add_text(tag=f'{spn}_hex')

                                # OPTIONS
                                labels = [label for label in spn_spec['discrete_values'].values()]
                                dpg.add_combo(tag=f'{pgn}_{spn}_combo', items=labels, user_data=(pgn, spn, spn_spec),
                                              default_value=J1939.get_label(signal_spec=spn_spec, value=0),
                                              label=f"SPN {spn}: {spn_spec['SPN Name']}",
                                              callback=self.discrete_spn_combo_changed)
                            self.discrete_spn_input_int_changed(str(spn), 0, (pgn, spn, spn_spec))

            with dpg.group(horizontal=True):
                dpg.add_text('Tx mode:')
                dpg.add_checkbox(tag=f'{pgn}_tx_mode', label='Continuous',
                                 user_data=pgn, callback=self.pgn_tx_mode_changed)
                dpg.add_spacer(width=10)
                dpg.add_input_int(label='ms', default_value=self.J1939[pgn]['transmission_rate_ms'],
                                  min_value=10, max_value=50000, min_clamped=True, max_clamped=True,
                                  step=10, step_fast=100, width=90,
                                  user_data=pgn, callback=self.pgn_tx_rate_changed)
                dpg.add_spacer(width=10)
                dpg.add_button(tag=f'{pgn}_tx_once', label='Tx Once',
                               user_data=pgn, callback=self.pgn_tx_once_invoked)

            dpg.add_spacer(height=10)

    def pgn_tx_mode_changed(self, sender, cont_tx, pgn):
        if cont_tx:
            dpg.hide_item(f'{pgn}_tx_once')
            self.transmitter.set_tx_mode_continuous(pgn)
        else:
            dpg.show_item(f'{pgn}_tx_once')
            self.transmitter.set_tx_mode_stop(pgn)

    def pgn_tx_once_invoked(self, sender, app_data, pgn):
        self.transmitter.set_tx_once(pgn)

    def pgn_tx_rate_changed(self, sender, tx_rate_ms, pgn):
        self.transmitter.modify_pgn_tx_rate(pgn, tx_rate_ms)

    def continuous_spn_slider_changed(self, sender, raw_value, pgn__spn__spn_spec):
        pgn, spn, signal_spec = pgn__spn__spn_spec
        self.transmitter.modify_pgn_data(pgn, spn, raw_value)

        decoded_value = J1939.decode(raw_value=raw_value,
                                     scale=signal_spec['scale'],
                                     offset=signal_spec['offset'])
        # :4.{len(str(J1939[pgn]['SPNs'][spn]['scale']).split('.')[1])}f
        dpg.set_value(f'{pgn}_{spn}_input', decoded_value)
        # dpg.set_item_user_data(f'{pgn}_{spn}_input', decoded_value)

        match signal_spec['length_bits']:
            case 1 | 2 | 3 | 4:
                nimble = f'{raw_value:X}'
                dpg.set_value(f'{spn}_hex', f'   [{nimble}]             ')
            case 5 | 6 | 7 | 8:
                byte = f'{raw_value:02X}'
                dpg.set_value(f'{spn}_hex', f'   [{byte}]            ')
            case 16:
                all_bytes = f'{raw_value:04X}'
                byte1 = all_bytes[2:4]
                byte2 = all_bytes[0:2]
                dpg.set_value(f'{spn}_hex', f'LSB[{byte1} {byte2}]MSB      ')
            case 32:
                all_bytes = f'{raw_value:08X}'
                byte1 = all_bytes[6:8]
                byte2 = all_bytes[4:6]
                byte3 = all_bytes[2:4]
                byte4 = all_bytes[0:2]
                dpg.set_value(f'{spn}_hex', f'LSB[{byte1} {byte2} {byte3} {byte4}]MSB')

    def continuous_spn_input_int_changed(self, sender, real_value, pgn__spn__spn_spec):
        pgn, spn, spn_spec = pgn__spn__spn_spec
        raw_value = J1939.encode(decoded_value=real_value, scale=spn_spec['scale'], offset=spn_spec['offset'])
        slider_tag = str(spn)
        dpg.set_value(slider_tag, raw_value)
        self.continuous_spn_slider_changed(slider_tag, raw_value, pgn__spn__spn_spec)

        #if new_value > old_value:
        #    ...
        #else:
        #    ...
        #    dpg.set_value(spn, dpg.get_value(spn) - 1)
        #    dpg.set_item_user_data(sender, new_value)

    def discrete_spn_input_int_changed(self, sender, raw_value, pgn__spn__spn_spec):
        pgn, spn, signal_spec = pgn__spn__spn_spec
        self.transmitter.modify_pgn_data(pgn, spn, raw_value)
        dpg.set_value(f'{pgn}_{spn}_combo', J1939.get_label(signal_spec=signal_spec, value=raw_value))

        match signal_spec['length_bits']:
            case 1 | 2 | 3 | 4 as length_bits:
                nibble = f'{raw_value:X}'
                dpg.set_value(f'{spn}_bin', f"{raw_value:0{length_bits}b}".ljust(10))
                dpg.set_value(f'{spn}_hex', f'    [{nibble}]            ')
            case 5 | 6 | 7 | 8 as length_bits:
                byte = f'{raw_value:02X}'
                dpg.set_value(f'{spn}_bin', f"{raw_value:0{length_bits}b}".ljust(10))
                dpg.set_value(f'{spn}_hex', f'   [{byte}]            ')

    def discrete_spn_combo_changed(self, sender, real_value, pgn__spn__spn_spec):
        pgn, spn, spn_spec = pgn__spn__spn_spec
        raw_value = J1939.get_label_value(signal_spec=spn_spec, label=real_value)
        input_int_widget = str(spn)
        dpg.set_value(input_int_widget, raw_value)
        self.discrete_spn_input_int_changed(input_int_widget, raw_value, pgn__spn__spn_spec)
